@@ -9,10 +9,10 @@
 #ifndef NASM_ASSEMBLE_H
 #define NASM_ASSEMBLE_H
 
-long insn_size (long segment, long offset, int bits,
-		insn *instruction, efunc error);
-long assemble (long segment, long offset, int bits,
-	       insn *instruction, struct ofmt *output, efunc error,
-	       ListGen *listgen);
+int32_t insn_size(int32_t segment, int32_t offset, int bits, uint32_t cp,
+               insn * instruction, efunc error);
+int32_t assemble(int32_t segment, int32_t offset, int bits, uint32_t cp,
+              insn * instruction, struct ofmt *output, efunc error,
+              ListGen * listgen);
 
 #endif
