@@ -8,16 +8,20 @@
  * initial version 27/iii/95 by Simon Tatham
  */
 
+#include "compiler.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <stddef.h>
 #include <string.h>
 #include <ctype.h>
+#include <inttypes.h>
 
 #include "nasm.h"
 #include "nasmlib.h"
 #include "eval.h"
 #include "labels.h"
+#include "float.h"
 
 #define TEMPEXPRS_DELTA 128
 #define TEMPEXPR_DELTA 8
@@ -40,14 +44,14 @@ static struct tokenval *tokval; /* The current token */
 static int i;                   /* The t_type of tokval */
 
 static void *scpriv;
-static loc_t *location;         /* Pointer to current line's segment,offset */
+static struct location *location;         /* Pointer to current line's segment,offset */
 static int *opflags;
 
 static struct eval_hints *hint;
 
 extern int in_abs_seg;          /* ABSOLUTE segment flag */
-extern long abs_seg;            /* ABSOLUTE segment */
-extern long abs_offset;         /* ABSOLUTE segment offset */
+extern int32_t abs_seg;            /* ABSOLUTE segment */
+extern int32_t abs_offset;         /* ABSOLUTE segment offset */
 
 /*
  * Unimportant cleanup is done to avoid confusing people who are trying
@@ -69,7 +73,7 @@ static void begintemp(void)
     tempexpr_size = ntempexpr = 0;
 }
 
-static void addtotemp(long type, long value)
+static void addtotemp(int32_t type, int64_t value)
 {
     while (ntempexpr >= tempexpr_size) {
         tempexpr_size += TEMPEXPR_DELTA;
@@ -116,7 +120,7 @@ static expr *add_vectors(expr * p, expr * q)
             addtotemp(p->type, p->value);
             lasttype = p++->type;
         } else {                /* *p and *q have same type */
-            long sum = p->value + q->value;
+            int64_t sum = p->value + q->value;
             if (sum)
                 addtotemp(p->type, sum);
             lasttype = p->type;
@@ -150,7 +154,7 @@ static expr *add_vectors(expr * p, expr * q)
  * multiplied. This allows [eax*1+ebx] to hint EBX rather than EAX
  * as the base register.
  */
-static expr *scalar_mult(expr * vect, long scalar, int affect_hints)
+static expr *scalar_mult(expr * vect, int64_t scalar, int affect_hints)
 {
     expr *p = vect;
 
@@ -166,7 +170,7 @@ static expr *scalar_mult(expr * vect, long scalar, int affect_hints)
     return vect;
 }
 
-static expr *scalarvect(long scalar)
+static expr *scalarvect(int64_t scalar)
 {
     begintemp();
     addtotemp(EXPR_SIMPLE, scalar);
@@ -187,7 +191,7 @@ static expr *unknown_expr(void)
  */
 static expr *segment_part(expr * e)
 {
-    long seg;
+    int32_t seg;
 
     if (is_unknown(e))
         return unknown_expr();
@@ -208,7 +212,7 @@ static expr *segment_part(expr * e)
               " is already a segment base");
         return NULL;
     } else {
-        long base = outfmt->segbase(seg + 1);
+        int32_t base = outfmt->segbase(seg + 1);
 
         begintemp();
         addtotemp((base == NO_SEG ? EXPR_UNKNOWN : EXPR_SEGBASE + base),
@@ -219,7 +223,7 @@ static expr *segment_part(expr * e)
 
 /*
  * Recursive-descent parser. Called with a single boolean operand,
- * which is TRUE if the evaluation is critical (i.e. unresolved
+ * which is true if the evaluation is critical (i.e. unresolved
  * symbols are an error condition). Must update the global `i' to
  * reflect the token after the parsed string. May return NULL.
  *
@@ -279,7 +283,7 @@ static expr *rexp0(int critical)
         if (is_just_unknown(e) || is_just_unknown(f))
             e = unknown_expr();
         else
-            e = scalarvect((long)(reloc_value(e) || reloc_value(f)));
+            e = scalarvect((int64_t)(reloc_value(e) || reloc_value(f)));
     }
     return e;
 }
@@ -306,7 +310,7 @@ static expr *rexp1(int critical)
         if (is_just_unknown(e) || is_just_unknown(f))
             e = unknown_expr();
         else
-            e = scalarvect((long)(!reloc_value(e) ^ !reloc_value(f)));
+            e = scalarvect((int64_t)(!reloc_value(e) ^ !reloc_value(f)));
     }
     return e;
 }
@@ -331,7 +335,7 @@ static expr *rexp2(int critical)
         if (is_just_unknown(e) || is_just_unknown(f))
             e = unknown_expr();
         else
-            e = scalarvect((long)(reloc_value(e) && reloc_value(f)));
+            e = scalarvect((int64_t)(reloc_value(e) && reloc_value(f)));
     }
     return e;
 }
@@ -339,7 +343,7 @@ static expr *rexp2(int critical)
 static expr *rexp3(int critical)
 {
     expr *e, *f;
-    long v;
+    int64_t v;
 
     e = expr0(critical);
     if (!e)
@@ -353,7 +357,7 @@ static expr *rexp3(int critical)
         if (!f)
             return NULL;
 
-        e = add_vectors(e, scalar_mult(f, -1L, FALSE));
+        e = add_vectors(e, scalar_mult(f, -1L, false));
 
         switch (j) {
         case TOKEN_EQ:
@@ -361,9 +365,9 @@ static expr *rexp3(int critical)
             if (is_unknown(e))
                 v = -1;         /* means unknown */
             else if (!is_really_simple(e) || reloc_value(e) != 0)
-                v = (j == TOKEN_NE);    /* unequal, so return TRUE if NE */
+                v = (j == TOKEN_NE);    /* unequal, so return true if NE */
             else
-                v = (j == TOKEN_EQ);    /* equal, so return TRUE if EQ */
+                v = (j == TOKEN_EQ);    /* equal, so return true if EQ */
             break;
         default:
             if (is_unknown(e))
@@ -498,7 +502,7 @@ static expr *expr3(int critical)
                 e = scalarvect(reloc_value(e) << reloc_value(f));
                 break;
             case TOKEN_SHR:
-                e = scalarvect(((unsigned long)reloc_value(e)) >>
+                e = scalarvect(((uint64_t)reloc_value(e)) >>
                                reloc_value(f));
                 break;
             }
@@ -524,7 +528,7 @@ static expr *expr4(int critical)
             e = add_vectors(e, f);
             break;
         case '-':
-            e = add_vectors(e, scalar_mult(f, -1L, FALSE));
+            e = add_vectors(e, scalar_mult(f, -1L, false));
             break;
         }
     }
@@ -558,9 +562,9 @@ static expr *expr5(int critical)
         switch (j) {
         case '*':
             if (is_simple(e))
-                e = scalar_mult(f, reloc_value(e), TRUE);
+                e = scalar_mult(f, reloc_value(e), true);
             else if (is_simple(f))
-                e = scalar_mult(e, reloc_value(f), TRUE);
+                e = scalar_mult(e, reloc_value(f), true);
             else if (is_just_unknown(e) && is_just_unknown(f))
                 e = unknown_expr();
             else {
@@ -573,51 +577,111 @@ static expr *expr5(int critical)
             if (is_just_unknown(e) || is_just_unknown(f))
                 e = unknown_expr();
             else
-                e = scalarvect(((unsigned long)reloc_value(e)) /
-                               ((unsigned long)reloc_value(f)));
+                e = scalarvect(((uint64_t)reloc_value(e)) /
+                               ((uint64_t)reloc_value(f)));
             break;
         case '%':
             if (is_just_unknown(e) || is_just_unknown(f))
                 e = unknown_expr();
             else
-                e = scalarvect(((unsigned long)reloc_value(e)) %
-                               ((unsigned long)reloc_value(f)));
+                e = scalarvect(((uint64_t)reloc_value(e)) %
+                               ((uint64_t)reloc_value(f)));
             break;
         case TOKEN_SDIV:
             if (is_just_unknown(e) || is_just_unknown(f))
                 e = unknown_expr();
             else
-                e = scalarvect(((signed long)reloc_value(e)) /
-                               ((signed long)reloc_value(f)));
+                e = scalarvect(((int64_t)reloc_value(e)) /
+                               ((int64_t)reloc_value(f)));
             break;
         case TOKEN_SMOD:
             if (is_just_unknown(e) || is_just_unknown(f))
                 e = unknown_expr();
             else
-                e = scalarvect(((signed long)reloc_value(e)) %
-                               ((signed long)reloc_value(f)));
+                e = scalarvect(((int64_t)reloc_value(e)) %
+                               ((int64_t)reloc_value(f)));
             break;
         }
     }
     return e;
 }
 
+static expr *eval_floatize(enum floatize type)
+{
+    uint8_t result[16], *p;	/* Up to 128 bits */
+    static const struct {
+	int bytes, start, len;
+    } formats[] = {
+	{  1, 0, 1 },		/* FLOAT_8 */
+	{  2, 0, 2 },		/* FLOAT_16 */
+	{  4, 0, 4 },		/* FLOAT_32 */
+	{  8, 0, 8 },		/* FLOAT_64 */
+	{ 10, 0, 8 },		/* FLOAT_80M */
+	{ 10, 8, 2 },		/* FLOAT_80E */
+	{ 16, 0, 8 },		/* FLOAT_128L */
+	{ 16, 8, 8 },		/* FLOAT_128H */
+    };
+    int sign = 1;
+    int64_t val;
+    int j;
+
+    i = scan(scpriv, tokval);
+    if (i != '(') {
+	error(ERR_NONFATAL, "expecting `('");
+	return NULL;
+    }
+    i = scan(scpriv, tokval);
+    if (i == '-' || i == '+') {
+	sign = (i == '-') ? -1 : 1;
+	i = scan(scpriv, tokval);
+    }
+    if (i != TOKEN_FLOAT) {
+	error(ERR_NONFATAL, "expecting floating-point number");
+	return NULL;
+    }
+    if (!float_const(tokval->t_charptr, sign, result,
+		     formats[type].bytes, error))
+	return NULL;
+    i = scan(scpriv, tokval);
+    if (i != ')') {
+	error(ERR_NONFATAL, "expecting `)'");
+	return NULL;
+    }
+
+    p = result+formats[type].start+formats[type].len;
+    val = 0;
+    for (j = formats[type].len; j; j--) {
+	p--;
+	val = (val << 8) + *p;
+    }
+
+    begintemp();
+    addtotemp(EXPR_SIMPLE, val);
+
+    i = scan(scpriv, tokval);
+    return finishtemp();
+}
+
 static expr *expr6(int critical)
 {
-    long type;
+    int32_t type;
     expr *e;
-    long label_seg, label_ofs;
+    int32_t label_seg, label_ofs;
 
-    if (i == '-') {
+    switch (i) {
+    case '-':
         i = scan(scpriv, tokval);
         e = expr6(critical);
         if (!e)
             return NULL;
-        return scalar_mult(e, -1L, FALSE);
-    } else if (i == '+') {
+        return scalar_mult(e, -1L, false);
+
+
+    case '+':
         i = scan(scpriv, tokval);
         return expr6(critical);
-    } else if (i == '~') {
+
+    case '~':
         i = scan(scpriv, tokval);
         e = expr6(critical);
         if (!e)
@@ -630,7 +694,22 @@ static expr *expr6(int critical)
             return NULL;
         }
         return scalarvect(~reloc_value(e));
-    } else if (i == TOKEN_SEG) {
+
+    case '!':
+        i = scan(scpriv, tokval);
+        e = expr6(critical);
+        if (!e)
+            return NULL;
+        if (is_just_unknown(e))
+            return unknown_expr();
+        else if (!is_simple(e)) {
+            error(ERR_NONFATAL, "`!' operator may only be applied to"
+                  " scalar values");
+            return NULL;
+        }
+        return scalarvect(!reloc_value(e));
+
+    case TOKEN_SEG:
         i = scan(scpriv, tokval);
         e = expr6(critical);
         if (!e)
@@ -643,7 +722,11 @@ static expr *expr6(int critical)
             return NULL;
         }
         return e;
-    } else if (i == '(') {
+
+    case TOKEN_FLOATIZE:
+	return eval_floatize(tokval->t_integer);
+
+    case '(':
         i = scan(scpriv, tokval);
         e = bexpr(critical);
         if (!e)
@@ -654,8 +737,12 @@ static expr *expr6(int critical)
         }
         i = scan(scpriv, tokval);
         return e;
-    } else if (i == TOKEN_NUM || i == TOKEN_REG || i == TOKEN_ID ||
-               i == TOKEN_HERE || i == TOKEN_BASE) {
+
+    case TOKEN_NUM:
+    case TOKEN_REG:
+    case TOKEN_ID:
+    case TOKEN_HERE:
+    case TOKEN_BASE:
         begintemp();
         switch (i) {
         case TOKEN_NUM:
@@ -719,14 +806,15 @@ static expr *expr6(int critical)
         }
         i = scan(scpriv, tokval);
         return finishtemp();
-    } else {
+
+    default:
         error(ERR_NONFATAL, "expression syntax error");
         return NULL;
     }
 }
 
 void eval_global_info(struct ofmt *output, lfunc lookup_label,
-                      loc_t * locp)
+                      struct location * locp)
 {
     outfmt = output;
     labelfunc = lookup_label;
@@ -761,7 +849,7 @@ expr *evaluate(scanner sc, void *scprivate, struct tokenval *tv,
     else
         i = tokval->t_type;
 
-    while (ntempexprs)          /* initialise temporary storage */
+    while (ntempexprs)          /* initialize temporary storage */
         nasm_free(tempexprs[--ntempexprs]);
 
     e = bexpr(critical);
@@ -774,13 +862,13 @@ expr *evaluate(scanner sc, void *scprivate, struct tokenval *tv,
         if (!f)
             return NULL;
     }
-    e = scalar_mult(e, 1L, FALSE);      /* strip far-absolute segment part */
+    e = scalar_mult(e, 1L, false);      /* strip far-absolute segment part */
     if (f) {
         expr *g;
         if (is_just_unknown(f))
             g = unknown_expr();
         else {
-            long value;
+            int64_t value;
             begintemp();
             if (!is_reloc(f)) {
                 error(ERR_NONFATAL, "invalid right-hand operand to WRT");
