@@ -11,22 +11,16 @@
 #ifndef NASM_NASM_H
 #define NASM_NASM_H
 
+#include "compiler.h"
+
 #include <stdio.h>
-#include "version.h"		       /* generated NASM version macros */
+#include <inttypes.h>
+#include "version.h"            /* generated NASM version macros */
+#include "nasmlib.h"
+#include "insnsi.h"		/* For enum opcode */
 
-#ifndef NULL
-#define NULL 0
-#endif
-
-#ifndef FALSE
-#define FALSE 0			       /* comes in handy */
-#endif
-#ifndef TRUE
-#define TRUE 1
-#endif
-
-#define NO_SEG -1L		       /* null segment value */
-#define SEG_ABS 0x40000000L	       /* mask for far-absolute segments */
+#define NO_SEG -1L              /* null segment value */
+#define SEG_ABS 0x40000000L     /* mask for far-absolute segments */
 
 #ifndef FILENAME_MAX
 #define FILENAME_MAX 256
@@ -57,48 +51,6 @@
  */
 struct ofmt;
 
-/*
- * -------------------------
- * Error reporting functions
- * -------------------------
- */
-
-/*
- * An error reporting function should look like this.
- */
-typedef void (*efunc) (int severity, const char *fmt, ...);
-
-/*
- * These are the error severity codes which get passed as the first
- * argument to an efunc.
- */
-
-#define ERR_DEBUG  	0x00000008	/* put out debugging message */
-#define ERR_WARNING	0x00000000	/* warn only: no further action */
-#define ERR_NONFATAL	0x00000001	/* terminate assembly after phase */
-#define ERR_FATAL	0x00000002	/* instantly fatal: exit with error */
-#define ERR_PANIC	0x00000003	/* internal error: panic instantly
-					* and dump core for reference */
-#define ERR_MASK	0x0000000F	/* mask off the above codes */
-#define ERR_NOFILE	0x00000010	/* don't give source file name/line */
-#define ERR_USAGE	0x00000020	/* print a usage message */
-#define ERR_PASS1	0x00000040	/* only print this error on pass one */
-
-/*
- * These codes define specific types of suppressible warning.
- */
-
-#define ERR_WARN_MASK	0x0000FF00	/* the mask for this feature */
-#define ERR_WARN_SHR  8		       /* how far to shift right */
-
-#define ERR_WARN_MNP	0x00000100	/* macro-num-parameters warning */
-#define ERR_WARN_MSR	0x00000200	/* macro self-reference */
-#define ERR_WARN_OL	0x00000300	/* orphan label (no colon, and
-					* alone on line) */
-#define ERR_WARN_NOV	0x00000400	/* numeric overflow */
-#define ERR_WARN_GNUELF	0x00000500      /* using GNU ELF extensions */
-#define ERR_WARN_MAX	5		/* the highest numbered one */
-
 /*
  * -----------------------
  * Other function typedefs
@@ -108,7 +60,7 @@ typedef void (*efunc) (int severity, const char *fmt, ...);
 /*
  * A label-lookup function should look like this.
  */
-typedef int (*lfunc) (char *label, long *segment, long *offset);
+typedef bool (*lfunc) (char *label, int32_t *segment, int32_t *offset);
 
 /*
  * And a label-definition function like this. The boolean parameter
@@ -116,16 +68,16 @@ typedef int (*lfunc) (char *label, long *segment, long *offset);
  * should affect the local-label system), or something odder like
  * an EQU or a segment-base symbol, which shouldn't.
  */
-typedef void (*ldfunc) (char *label, long segment, long offset, char *special,
-			int is_norm, int isextrn, struct ofmt *ofmt,
-			efunc error);
+typedef void (*ldfunc) (char *label, int32_t segment, int32_t offset,
+                        char *special, bool is_norm, bool isextrn,
+                        struct ofmt * ofmt, efunc error);
 
 /*
  * List-file generators should look like this:
  */
 typedef struct {
     /*
-     * Called to initialise the listing file generator. Before this
+     * Called to initialize the listing file generator. Before this
      * is called, the other routines will silently do nothing when
      * called. The `char *' parameter is the file name to write the
      * listing to.
@@ -148,7 +100,7 @@ typedef struct {
      * work with when doing things like uplevel(LIST_TIMES) or
      * uplevel(LIST_INCBIN).
      */
-    void (*output) (long, const void *, unsigned long);
+    void (*output) (int32_t, const void *, uint32_t);
 
     /*
      * Called to send a text line to the listing generator. The
@@ -190,37 +142,49 @@ typedef struct {
  */
 struct tokenval {
     int t_type;
-    long t_integer, t_inttwo;
+    int64_t t_integer, t_inttwo;
     char *t_charptr;
 };
-typedef int (*scanner) (void *private_data, struct tokenval *tv);
+typedef int (*scanner) (void *private_data, struct tokenval * tv);
 
 /*
  * Token types returned by the scanner, in addition to ordinary
  * ASCII character values, and zero for end-of-string.
  */
-enum {				       /* token types, other than chars */
-    TOKEN_INVALID = -1,		       /* a placeholder value */
-    TOKEN_EOS = 0,		       /* end of string */
-    TOKEN_EQ = '=', TOKEN_GT = '>', TOKEN_LT = '<',   /* aliases */
-    TOKEN_ID = 256, TOKEN_NUM, TOKEN_REG, TOKEN_INSN,  /* major token types */
-    TOKEN_ERRNUM,		       /* numeric constant with error in */
-    TOKEN_HERE, TOKEN_BASE,	       /* $ and $$ */
-    TOKEN_SPECIAL,		       /* BYTE, WORD, DWORD, FAR, NEAR, etc */
-    TOKEN_PREFIX,		       /* A32, O16, LOCK, REPNZ, TIMES, etc */
-    TOKEN_SHL, TOKEN_SHR,	       /* << and >> */
-    TOKEN_SDIV, TOKEN_SMOD,	       /* // and %% */
-    TOKEN_GE, TOKEN_LE, TOKEN_NE,      /* >=, <= and <> (!= is same as <>) */
-    TOKEN_DBL_AND, TOKEN_DBL_OR, TOKEN_DBL_XOR,   /* &&, || and ^^ */
-    TOKEN_SEG, TOKEN_WRT,	       /* SEG and WRT */
-    TOKEN_FLOAT			       /* floating-point constant */
+enum {                          /* token types, other than chars */
+    TOKEN_INVALID = -1,         /* a placeholder value */
+    TOKEN_EOS = 0,              /* end of string */
+    TOKEN_EQ = '=', TOKEN_GT = '>', TOKEN_LT = '<',     /* aliases */
+    TOKEN_ID = 256, TOKEN_NUM, TOKEN_REG, TOKEN_INSN,   /* major token types */
+    TOKEN_ERRNUM,               /* numeric constant with error in */
+    TOKEN_HERE, TOKEN_BASE,     /* $ and $$ */
+    TOKEN_SPECIAL,              /* BYTE, WORD, DWORD, QWORD, FAR, NEAR, etc */
+    TOKEN_PREFIX,               /* A32, O16, LOCK, REPNZ, TIMES, etc */
+    TOKEN_SHL, TOKEN_SHR,       /* << and >> */
+    TOKEN_SDIV, TOKEN_SMOD,     /* // and %% */
+    TOKEN_GE, TOKEN_LE, TOKEN_NE,       /* >=, <= and <> (!= is same as <>) */
+    TOKEN_DBL_AND, TOKEN_DBL_OR, TOKEN_DBL_XOR, /* &&, || and ^^ */
+    TOKEN_SEG, TOKEN_WRT,       /* SEG and WRT */
+    TOKEN_FLOAT,                /* floating-point constant */
+    TOKEN_FLOATIZE,		/* __floatX__ */
 };
 
-typedef struct {
-    long segment;
-    long offset;
-    int  known;
-} loc_t;
+enum floatize {
+    FLOAT_8,
+    FLOAT_16,
+    FLOAT_32,
+    FLOAT_64,
+    FLOAT_80M,
+    FLOAT_80E,
+    FLOAT_128L,
+    FLOAT_128H,
+};
+
+struct location {
+    int64_t offset;
+    int32_t segment;
+    int known;
+};
 
 /*
  * Expression-evaluator datatype. Expressions, within the
@@ -235,17 +199,29 @@ typedef struct {
  * `value' field of zero is insignificant.
  */
 typedef struct {
-    long type;			       /* a register, or EXPR_xxx */
-    long value;			       /* must be >= 32 bits */
+    int32_t type;                  /* a register, or EXPR_xxx */
+    int64_t value;                 /* must be >= 32 bits */
 } expr;
 
+/*
+ * Library routines to manipulate expression data types.
+ */
+int is_reloc(expr *);
+int is_simple(expr *);
+int is_really_simple(expr *);
+int is_unknown(expr *);
+int is_just_unknown(expr *);
+int64_t reloc_value(expr *);
+int32_t reloc_seg(expr *);
+int32_t reloc_wrt(expr *);
+
 /*
  * The evaluator can also return hints about which of two registers
  * used in an expression should be the base register. See also the
  * `operand' structure.
  */
 struct eval_hints {
-    int base;
+    int64_t base;
     int type;
 };
 
@@ -256,7 +232,7 @@ struct eval_hints {
  * it will start by calling the scanner.
  *
  * If a forward reference happens during evaluation, the evaluator
- * must set `*fwref' to TRUE if `fwref' is non-NULL.
+ * must set `*fwref' to true if `fwref' is non-NULL.
  *
  * `critical' is non-zero if the expression may not contain forward
  * references. The evaluator will report its own error if this
@@ -274,27 +250,24 @@ struct eval_hints {
  * the base register in complex effective addresses.
  */
 #define CRITICAL 0x100
-typedef expr *(*evalfunc) (scanner sc, void *scprivate, struct tokenval *tv,
-			   int *fwref, int critical, efunc error,
-			   struct eval_hints *hints);
+typedef expr *(*evalfunc) (scanner sc, void *scprivate,
+                           struct tokenval * tv, int *fwref, int critical,
+                           efunc error, struct eval_hints * hints);
 
 /*
- * Special values for expr->type. ASSUMPTION MADE HERE: the number
- * of distinct register names (i.e. possible "type" fields for an
- * expr structure) does not exceed 124 (EXPR_REG_START through
- * EXPR_REG_END).
+ * Special values for expr->type.  These come after EXPR_REG_END
+ * as defined in regs.h.
  */
-#define EXPR_REG_START 1
-#define EXPR_REG_END 124
-#define EXPR_UNKNOWN 125L	       /* for forward references */
-#define EXPR_SIMPLE 126L
-#define EXPR_WRT 127L
-#define EXPR_SEGBASE 128L
+
+#define EXPR_UNKNOWN	(EXPR_REG_END+1) /* forward references */
+#define EXPR_SIMPLE	(EXPR_REG_END+2)
+#define EXPR_WRT	(EXPR_REG_END+3)
+#define EXPR_SEGBASE	(EXPR_REG_END+4)
 
 /*
  * Preprocessors ought to look like this:
  */
-typedef struct {
+typedef struct preproc_ops {
     /*
      * Called at the start of a pass; given a file name, the number
      * of the pass, an error reporting function, an evaluator
@@ -315,6 +288,8 @@ typedef struct {
     void (*cleanup) (int);
 } Preproc;
 
+extern Preproc nasmpp;
+
 /*
  * ----------------------------------------------------------------
  * Some lexical properties of the NASM source language, included
@@ -327,7 +302,7 @@ typedef struct {
  * matches any character that may appear at places other than the start of an
  * identifier. E.g. a period may only appear at the start of an identifier
  * (for local labels), whereas a number may appear anywhere *but* at the
- * start. 
+ * start.
  */
 
 #define isidstart(c) ( isalpha(c) || (c)=='_' || (c)=='.' || (c)=='?' \
@@ -338,7 +313,7 @@ typedef struct {
 /* Ditto for numeric constants. */
 
 #define isnumstart(c)  ( isdigit(c) || (c)=='$' )
-#define isnumchar(c)   ( isalnum(c) )
+#define isnumchar(c)   ( isalnum(c) || (c)=='_' )
 
 /* This returns the numeric value of a given 'digit'. */
 
@@ -365,149 +340,295 @@ enum {
  * instruction is a special operand type, whereas AX in other
  * contexts is just another 16-bit register. (Also, consider CL in
  * shift instructions, DX in OUT, etc.)
+ *
+ * The basic concept here is that
+ *    (class & ~operand) == 0
+ *
+ * if and only if "operand" belongs to class type "class".
+ *
+ * The bits are assigned as follows:
+ *
+ * Bits 0-7, 29: sizes
+ *  0:  8 bits (BYTE)
+ *  1: 16 bits (WORD)
+ *  2: 32 bits (DWORD)
+ *  3: 64 bits (QWORD)
+ *  4: 80 bits (TWORD)
+ *  5: FAR
+ *  6: NEAR
+ *  7: SHORT
+ * 29: 128 bits (OWORD)
+ *
+ * Bits 8-11 modifiers
+ *  8: TO
+ *  9: COLON
+ * 10: STRICT
+ * 11: (reserved)
+ *
+ * Bits 12-15: type of operand
+ * 12: REGISTER
+ * 13: IMMEDIATE
+ * 14: MEMORY (always has REGMEM attribute as well)
+ * 15: REGMEM (valid EA operand)
+ *
+ * Bits 16-19: subclasses
+ * With REG_CDT:
+ * 16: REG_CREG (CRx)
+ * 17: REG_DREG (DRx)
+ * 18: REG_TREG (TRx)
+
+ * With REG_GPR:
+ * 16: REG_ACCUM (AL, AX, EAX, RAX)
+ * 17: REG_COUNT (CL, CX, ECX, RCX)
+ * 18: REG_DATA  (DL, DX, EDX, RDX)
+ * 19: REG_HIGH  (AH, CH, DH, BH)
+ *
+ * With REG_SREG:
+ * 16: REG_CS
+ * 17: REG_DESS (DS, ES, SS)
+ * 18: REG_FSGS
+ * 19: REG_SEG67
+ *
+ * With FPUREG:
+ * 16: FPU0
+ *
+ * With XMMREG:
+ * 16: XMM0
+ *
+ * With MEMORY:
+ * 16: MEM_OFFS (this is a simple offset)
+ * 17: IP_REL (IP-relative offset)
+ *
+ * With IMMEDIATE:
+ * 16: UNITY (1)
+ * 17: BYTENESS (-128..127)
+ *
+ * Bits 20-26: register classes
+ * 20: REG_CDT (CRx, DRx, TRx)
+ * 21: RM_GPR (REG_GPR) (integer register)
+ * 22: REG_SREG
+ * 23: IP_REG (RIP or EIP) [unused]
+ * 24: FPUREG
+ * 25: RM_MMX (MMXREG)
+ * 26: RM_XMM (XMMREG)
+ *
+ * Bits 27-29 & 31 are currently unallocated.
+ *
+ * 30: SAME_AS
+ * Special flag only used in instruction patterns; means this operand
+ * has to be identical to another operand.  Currently only supported
+ * for registers.
  */
 
-/* size, and other attributes, of the operand */
-#define BITS8     0x00000001L
-#define BITS16    0x00000002L
-#define BITS32    0x00000004L
-#define BITS64    0x00000008L	       /* FPU only */
-#define BITS80    0x00000010L	       /* FPU only */
-#define FAR       0x00000020L	       /* grotty: this means 16:16 or */
-				       /* 16:32, like in CALL/JMP */
-#define NEAR      0x00000040L
-#define SHORT     0x00000080L	       /* and this means what it says :) */
+typedef uint32_t opflags_t;
 
-#define SIZE_MASK 0x000000FFL	       /* all the size attributes */
-#define NON_SIZE  (~SIZE_MASK)
+/* Size, and other attributes, of the operand */
+#define BITS8		0x00000001L
+#define BITS16		0x00000002L
+#define BITS32		0x00000004L
+#define BITS64		0x00000008L   /* x64 and FPU only */
+#define BITS80		0x00000010L   /* FPU only */
+#define BITS128		0x20000000L
+#define FAR		0x00000020L   /* grotty: this means 16:16 or */
+                                       /* 16:32, like in CALL/JMP */
+#define NEAR		0x00000040L
+#define SHORT		0x00000080L   /* and this means what it says :) */
 
-#define TO        0x00000100L          /* reverse effect in FADD, FSUB &c */
-#define COLON     0x00000200L	       /* operand is followed by a colon */
-#define STRICT    0x00000400L	       /* do not optimize this operand */
+#define SIZE_MASK	0x200000FFL   /* all the size attributes */
 
-/* type of operand: memory reference, register, etc. */
-#define MEMORY    0x00204000L
-#define REGISTER  0x00001000L	       /* register number in 'basereg' */
-#define IMMEDIATE 0x00002000L
+/* Modifiers */
+#define MODIFIER_MASK	0x00000f00L
+#define TO		0x00000100L   /* reverse effect in FADD, FSUB &c */
+#define COLON		0x00000200L   /* operand is followed by a colon */
+#define STRICT		0x00000400L   /* do not optimize this operand */
 
-#define REGMEM    0x00200000L	       /* for r/m, ie EA, operands */
-#define REGNORM   0x00201000L	       /* 'normal' reg, qualifies as EA */
-#define REG8      0x00201001L
-#define REG16     0x00201002L
-#define REG32     0x00201004L
-#define MMXREG    0x00201008L	       /* MMX registers */
-#define XMMREG    0x00201010L          /* XMM Katmai reg */
-#define FPUREG    0x01000000L	       /* floating point stack registers */
-#define FPU0      0x01000800L	       /* FPU stack register zero */
+/* Type of operand: memory reference, register, etc. */
+#define OPTYPE_MASK	0x0000f000L
+#define REGISTER	0x00001000L   /* register number in 'basereg' */
+#define IMMEDIATE	0x00002000L
+#define MEMORY		0x0000c000L
+#define REGMEM		0x00008000L   /* for r/m, ie EA, operands */
 
-/* special register operands: these may be treated differently */
-#define REG_SMASK 0x00070000L	       /* a mask for the following */
-#define REG_ACCUM 0x00211000L	       /* accumulator: AL, AX or EAX */
-#define REG_AL    0x00211001L	       /* REG_ACCUM | BITSxx */
-#define REG_AX    0x00211002L	       /* ditto */
-#define REG_EAX   0x00211004L	       /* and again */
-#define REG_COUNT 0x00221000L	       /* counter: CL, CX or ECX */
-#define REG_CL    0x00221001L	       /* REG_COUNT | BITSxx */
-#define REG_CX    0x00221002L	       /* ditto */
-#define REG_ECX   0x00221004L	       /* another one */
-#define REG_DL    0x00241001L
-#define REG_DX    0x00241002L
-#define REG_EDX   0x00241004L
-#define REG_SREG  0x00081002L	       /* any segment register */
-#define REG_CS    0x01081002L	       /* CS */
-#define REG_DESS  0x02081002L	       /* DS, ES, SS (non-CS 86 registers) */
-#define REG_FSGS  0x04081002L	       /* FS, GS (386 extended registers) */
-#define REG_SEG67 0x08081002L          /* Non-implemented segment registers */
-#define REG_CDT   0x00101004L	       /* CRn, DRn and TRn */
-#define REG_CREG  0x08101004L	       /* CRn */
-#define REG_DREG  0x10101004L	       /* DRn */
-#define REG_TREG  0x20101004L	       /* TRn */
+/* Register classes */
+#define REG_EA		0x00009000L   /* 'normal' reg, qualifies as EA */
+#define RM_GPR		0x00208000L   /* integer operand */
+#define REG_GPR		0x00209000L   /* integer register */
+#define REG8		0x00209001L   /*  8-bit GPR  */
+#define REG16		0x00209002L   /* 16-bit GPR */
+#define REG32		0x00209004L   /* 32-bit GPR */
+#define REG64		0x00209008L   /* 64-bit GPR */
+#define IP_REG		0x00801000L   /* RIP or EIP register */
+#define RIPREG		0x00801008L   /* RIP */
+#define EIPREG		0x00801004L   /* EIP */
+#define FPUREG		0x01001000L   /* floating point stack registers */
+#define FPU0		0x01011000L   /* FPU stack register zero */
+#define RM_MMX		0x02008000L   /* MMX operand */
+#define MMXREG		0x02009000L   /* MMX register */
+#define RM_XMM		0x04008000L   /* XMM (SSE) operand */
+#define XMMREG		0x04009000L   /* XMM (SSE) register */
+#define XMM0		0x04019000L   /* XMM register zero */
+#define REG_CDT		0x00101004L   /* CRn, DRn and TRn */
+#define REG_CREG	0x00111004L   /* CRn */
+#define REG_DREG	0x00121004L   /* DRn */
+#define REG_TREG	0x00141004L   /* TRn */
+#define REG_SREG	0x00401002L   /* any segment register */
+#define REG_CS		0x00411002L   /* CS */
+#define REG_DESS	0x00421002L   /* DS, ES, SS */
+#define REG_FSGS	0x00441002L   /* FS, GS */
+#define REG_SEG67	0x00481002L   /* Unimplemented segment registers */
 
-/* special type of EA */
-#define MEM_OFFS  0x00604000L	       /* simple [address] offset */
+#define REG_RIP		0x00801008L   /* RIP relative addressing */
+#define REG_EIP		0x00801004L   /* EIP relative addressing */
+
+/* Special GPRs */
+#define REG_SMASK	0x000f0000L   /* a mask for the following */
+#define REG_ACCUM	0x00219000L   /* accumulator: AL, AX, EAX, RAX */
+#define REG_AL		0x00219001L
+#define REG_AX		0x00219002L
+#define REG_EAX		0x00219004L
+#define REG_RAX		0x00219008L
+#define REG_COUNT	0x00229000L   /* counter: CL, CX, ECX, RCX */
+#define REG_CL		0x00229001L
+#define REG_CX		0x00229002L
+#define REG_ECX		0x00229004L
+#define REG_RCX		0x00229008L
+#define REG_DL		0x00249001L   /* data: DL, DX, EDX, RDX */
+#define REG_DX		0x00249002L
+#define REG_EDX		0x00249004L
+#define REG_RDX		0x00249008L
+#define REG_HIGH	0x00289001L   /* high regs: AH, CH, DH, BH */
+
+/* special types of EAs */
+#define MEM_OFFS	0x0001c000L   /* simple [address] offset - absolute! */
+#define IP_REL		0x0002c000L   /* IP-relative offset */
+
+/* memory which matches any type of r/m operand */
+#define MEMORY_ANY	(MEMORY|RM_GPR|RM_MMX|RM_XMM)
 
 /* special type of immediate operand */
-#define ONENESS   0x00800000L          /* so UNITY == IMMEDIATE | ONENESS */
-#define UNITY     0x00802000L	       /* for shift/rotate instructions */
-#define BYTENESS  0x40000000L          /* so SBYTE == IMMEDIATE | BYTENESS */
-#define SBYTE 	  0x40002000L	       /* for op r16/32,immediate instrs. */
-		
+#define UNITY		0x00012000L   /* for shift/rotate instructions */
+#define SBYTE		0x00022000L   /* for op r16/32,immediate instrs. */
+
+/* special flags */
+#define SAME_AS		0x40000000L
+
 /* Register names automatically generated from regs.dat */
 #include "regs.h"
 
-enum {				       /* condition code names */
+enum ccode {			/* condition code names */
     C_A, C_AE, C_B, C_BE, C_C, C_E, C_G, C_GE, C_L, C_LE, C_NA, C_NAE,
     C_NB, C_NBE, C_NC, C_NE, C_NG, C_NGE, C_NL, C_NLE, C_NO, C_NP,
-    C_NS, C_NZ, C_O, C_P, C_PE, C_PO, C_S, C_Z
+    C_NS, C_NZ, C_O, C_P, C_PE, C_PO, C_S, C_Z,
+    C_none = -1
 };
 
+/*
+ * REX flags
+ */
+#define REX_OC		0x0200	/* DREX suffix has the OC0 bit set */
+#define REX_D		0x0100	/* Instruction uses DREX instead of REX */
+#define REX_H		0x80	/* High register present, REX forbidden */
+#define REX_P		0x40	/* REX prefix present/required */
+#define REX_L		0x20	/* Use LOCK prefix instead of REX.R */
+#define REX_W		0x08	/* 64-bit operand size */
+#define REX_R		0x04	/* ModRM reg extension */
+#define REX_X		0x02	/* SIB index extension */
+#define REX_B		0x01	/* ModRM r/m extension */
+#define REX_REAL	0x4f	/* Actual REX prefix bits */
+
 /*
  * Note that because segment registers may be used as instruction
  * prefixes, we must ensure the enumerations for prefixes and
  * register names do not overlap.
  */
-enum {				       /* instruction prefixes */
+enum prefixes {			/* instruction prefixes */
+    P_none = 0,
     PREFIX_ENUM_START = REG_ENUM_LIMIT,
-    P_A16 = PREFIX_ENUM_START, P_A32, P_LOCK, P_O16, P_O32, P_REP, P_REPE,
-    P_REPNE, P_REPNZ, P_REPZ, P_TIMES
+    P_A16 = PREFIX_ENUM_START, P_A32, P_A64, P_ASP,
+    P_LOCK, P_O16, P_O32, P_O64, P_OSP,
+    P_REP, P_REPE, P_REPNE, P_REPNZ, P_REPZ, P_TIMES,
+    PREFIX_ENUM_LIMIT
 };
 
-enum {				       /* extended operand types */
+enum {                          /* extended operand types */
     EOT_NOTHING, EOT_DB_STRING, EOT_DB_NUMBER
 };
 
-enum {				       /* special EA flags */
-    EAF_BYTEOFFS = 1,		       /* force offset part to byte size */
-    EAF_WORDOFFS = 2,		       /* force offset part to [d]word size */
-    EAF_TIMESTWO = 4		       /* really do EAX*2 not EAX+EAX */
+enum {                          /* special EA flags */
+    EAF_BYTEOFFS =  1,          /* force offset part to byte size */
+    EAF_WORDOFFS =  2,          /* force offset part to [d]word size */
+    EAF_TIMESTWO =  4,          /* really do EAX*2 not EAX+EAX */
+    EAF_REL	 =  8,		/* IP-relative addressing */
+    EAF_ABS      = 16,		/* non-IP-relative addressing */
+    EAF_FSGS	 = 32		/* fs/gs segment override present */
 };
 
-enum {				       /* values for `hinttype' */
-    EAH_NOHINT = 0,		       /* no hint at all - our discretion */
-    EAH_MAKEBASE = 1,		       /* try to make given reg the base */
-    EAH_NOTBASE = 2		       /* try _not_ to make reg the base */
+enum eval_hint {                /* values for `hinttype' */
+    EAH_NOHINT   = 0,           /* no hint at all - our discretion */
+    EAH_MAKEBASE = 1,           /* try to make given reg the base */
+    EAH_NOTBASE  = 2            /* try _not_ to make reg the base */
 };
 
-typedef struct {		       /* operand to an instruction */
-    long type;			       /* type of operand */
-    int addr_size;		       /* 0 means default; 16; 32 */
-    int basereg, indexreg, scale;      /* registers and scale involved */
-    int hintbase, hinttype;	       /* hint as to real base register */
-    long segment;		       /* immediate segment, if needed */
-    long offset;		       /* any immediate number */
-    long wrt;			       /* segment base it's relative to */
-    int eaflags;		       /* special EA flags */
-    int opflags;		       /* see OPFLAG_* defines below */
+typedef struct operand {	/* operand to an instruction */
+    int32_t type;               /* type of operand */
+    int disp_size;              /* 0 means default; 16; 32; 64 */
+    enum reg_enum basereg, indexreg; /* address registers */
+    int scale;			/* index scale */
+    int hintbase;
+    enum eval_hint hinttype;    /* hint as to real base register */
+    int32_t segment;            /* immediate segment, if needed */
+    int64_t offset;             /* any immediate number */
+    int32_t wrt;                /* segment base it's relative to */
+    int eaflags;                /* special EA flags */
+    int opflags;                /* see OPFLAG_* defines below */
 } operand;
 
-#define OPFLAG_FORWARD		1      /* operand is a forward reference */
-#define OPFLAG_EXTERN		2      /* operand is an external reference */
+#define OPFLAG_FORWARD		1       /* operand is a forward reference */
+#define OPFLAG_EXTERN		2       /* operand is an external reference */
 
-typedef struct extop {		       /* extended operand */
-    struct extop *next;		       /* linked list */
-    long type;			       /* defined above */
-    char *stringval;		       /* if it's a string, then here it is */
-    int stringlen;		       /* ... and here's how long it is */
-    long segment;		       /* if it's a number/address, then... */
-    long offset;		       /* ... it's given here ... */
-    long wrt;			       /* ... and here */
+typedef struct extop {          /* extended operand */
+    struct extop *next;         /* linked list */
+    int32_t type;               /* defined above */
+    char *stringval;          /* if it's a string, then here it is */
+    int stringlen;              /* ... and here's how long it is */
+    int32_t segment;            /* if it's a number/address, then... */
+    int64_t offset;             /* ... it's given here ... */
+    int32_t wrt;                /* ... and here */
 } extop;
 
-#define MAXPREFIX 4
+/* Prefix positions: each type of prefix goes in a specific slot.
+   This affects the final ordering of the assembled output, which
+   shouldn't matter to the processor, but if you have stylistic
+   preferences, you can change this.  REX prefixes are handled
+   differently for the time being.
 
-typedef struct {		       /* an instruction itself */
-    char *label;		       /* the label defined, or NULL */
-    int prefixes[MAXPREFIX];	       /* instruction prefixes, if any */
-    int nprefix;		       /* number of entries in above */
-    int opcode;			       /* the opcode - not just the string */
-    int condition;		       /* the condition code, if Jcc/SETcc */
-    int operands;		       /* how many operands? 0-3 
-                                        * (more if db et al) */
-    operand oprs[3];	   	       /* the operands, defined as above */
-    extop *eops;		       /* extended operands */
-    int eops_float;                    /* true if DD and floating */
-    long times;			       /* repeat count (TIMES prefix) */
-    int forw_ref;		       /* is there a forward reference? */
+   Note that LOCK and REP are in the same slot.  This is
+   an x86 architectural constraint. */
+enum prefix_pos {
+    PPS_LREP,			/* Lock or REP prefix */
+    PPS_SEG,			/* Segment override prefix */
+    PPS_OSIZE,			/* Operand size prefix */
+    PPS_ASIZE,			/* Address size prefix */
+    MAXPREFIX			/* Total number of prefix slots */
+};
+
+#define MAX_OPERANDS 4
+
+typedef struct insn {		/* an instruction itself */
+    char *label;		/* the label defined, or NULL */
+    enum prefixes prefixes[MAXPREFIX]; /* instruction prefixes, if any */
+    enum opcode opcode;         /* the opcode - not just the string */
+    enum ccode condition;       /* the condition code, if Jcc/SETcc */
+    int operands;               /* how many operands? 0-3
+                                 * (more if db et al) */
+    int addr_size;		/* address size */
+    operand oprs[MAX_OPERANDS]; /* the operands, defined as above */
+    extop *eops;                /* extended operands */
+    int eops_float;             /* true if DD and floating */
+    int32_t times;              /* repeat count (TIMES prefix) */
+    int forw_ref;               /* is there a forward reference? */
+    int rex;			/* Special REX Prefix */
+    int drexdst;		/* Destination register for DREX suffix */
 } insn;
 
 enum geninfo { GI_SWITCH };
@@ -530,9 +651,10 @@ struct ofmt {
      */
     const char *shortname;
 
+
     /*
      * this is reserved for out module specific help.
-     * It is set to NULL in all the out modules but is not implemented
+     * It is set to NULL in all the out modules and is not implemented
      * in the main program
      */
     const char *helpstring;
@@ -567,7 +689,7 @@ struct ofmt {
      * to the label manager and expression evaluator if necessary.
      * It also gives it a chance to do other initialisation.
      */
-    void (*init) (FILE *fp, efunc error, ldfunc ldef, evalfunc eval);
+    void (*init) (FILE * fp, efunc error, ldfunc ldef, evalfunc eval);
 
     /*
      * This procedure is called to pass generic information to the
@@ -576,7 +698,7 @@ struct ofmt {
      * and the second parameter gives the value.  This function returns
      * 1 if recognized, 0 if unrecognized
      */
-    int (*setinfo)(enum geninfo type, char **string);
+    int (*setinfo) (enum geninfo type, char **string);
 
     /*
      * This procedure is called by assemble() to write actual
@@ -587,8 +709,8 @@ struct ofmt {
      * The `type' argument specifies the type of output data, and
      * usually the size as well: its contents are described below.
      */
-    void (*output) (long segto, const void *data, unsigned long type,
-		    long segment, long wrt);
+    void (*output) (int32_t segto, const void *data, uint32_t type,
+                    int32_t segment, int32_t wrt);
 
     /*
      * This procedure is called once for every symbol defined in
@@ -618,8 +740,8 @@ struct ofmt {
      * been an EXTERN, a COMMON or a GLOBAL. The distinction should
      * be obvious to the output format from the other parameters.
      */
-    void (*symdef) (char *name, long segment, long offset, int is_global,
-		    char *special);
+    void (*symdef) (char *name, int32_t segment, int32_t offset,
+		    int is_global, char *special);
 
     /*
      * This procedure is called when the source code requests a
@@ -636,7 +758,7 @@ struct ofmt {
      * the segment, by setting `*bits' to 16 or 32. Or, if it
      * doesn't wish to define a default, it can leave `bits' alone.
      */
-    long (*section) (char *name, int pass, int *bits);
+    int32_t (*section) (char *name, int pass, int *bits);
 
     /*
      * This procedure is called to modify the segment base values
@@ -651,7 +773,7 @@ struct ofmt {
      * responsible for throwing an error condition if that occurs
      * in pass two or in a critical expression.
      */
-    long (*segbase) (long segment);
+    int32_t (*segbase) (int32_t segment);
 
     /*
      * This procedure is called to allow the output driver to
@@ -676,7 +798,7 @@ struct ofmt {
      * should return its preferred name for the output file in
      * `outname', if outname[0] is not '\0', and do nothing to
      * `outname' otherwise. Since it is called before the driver is
-     * properly initialised, it has to be passed its error handler
+     * properly initialized, it has to be passed its error handler
      * separately.
      *
      * This procedure may also take its own copy of the input file
@@ -713,9 +835,9 @@ struct ofmt {
  * Also OUT_RESERVE denotes reservation of N bytes of BSS space,
  * and the contents of the "data" parameter is irrelevant.
  *
- * The "data" parameter for the output function points to a "long",
+ * The "data" parameter for the output function points to a "int32_t",
  * containing the address in question, unless the type is
- * OUT_RAWDATA, in which case it points to an "unsigned char"
+ * OUT_RAWDATA, in which case it points to an "uint8_t"
  * array.
  */
 #define OUT_RAWDATA 0x00000000UL
@@ -734,7 +856,7 @@ struct ofmt {
  */
 
 struct dfmt {
-    
+
     /*
      * This is a short (one-liner) description of the type of
      * output generated by the driver.
@@ -746,19 +868,18 @@ struct dfmt {
      */
     const char *shortname;
 
-
     /*
-     * init - called initially to set up local pointer to object format, 
+     * init - called initially to set up local pointer to object format,
      * void pointer to implementation defined data, file pointer (which
      * probably won't be used, but who knows?), and error function.
      */
-    void (*init) (struct ofmt * of, void * id, FILE * fp, efunc error);
+    void (*init) (struct ofmt * of, void *id, FILE * fp, efunc error);
 
     /*
      * linenum - called any time there is output with a change of
      * line number or file.
      */
-    void (*linenum) (const char * filename, long linenumber, long segto);
+    void (*linenum) (const char *filename, int32_t linenumber, int32_t segto);
 
     /*
      * debug_deflabel - called whenever a label is defined. Parameters
@@ -766,24 +887,24 @@ struct dfmt {
      * would be called before the output format version.
      */
 
-    void (*debug_deflabel) (char * name, long segment, long offset,
-                            int is_global, char * special);
+    void (*debug_deflabel) (char *name, int32_t segment, int32_t offset,
+                            int is_global, char *special);
     /*
      * debug_directive - called whenever a DEBUG directive other than 'LINE'
      * is encountered. 'directive' contains the first parameter to the
      * DEBUG directive, and params contains the rest. For example,
      * 'DEBUG VAR _somevar:int' would translate to a call to this
-     * function with 'directive' equal to "VAR" and 'params' equal to 
+     * function with 'directive' equal to "VAR" and 'params' equal to
      * "_somevar:int".
      */
-    void (*debug_directive) (const char * directive, const char * params);
+    void (*debug_directive) (const char *directive, const char *params);
 
     /*
      * typevalue - called whenever the assembler wishes to register a type
      * for the last defined label.  This routine MUST detect if a type was
      * already registered and not re-register it.
      */
-    void (*debug_typevalue) (long type);
+    void (*debug_typevalue) (int32_t type);
 
     /*
      * debug_output - called whenever output is required
@@ -814,6 +935,7 @@ struct dfmt {
 #define TY_FLOAT   0x28
 #define TY_QWORD   0x30
 #define TY_TBYTE   0x38
+#define TY_OWORD   0x40
 #define TY_COMMON  0xE0
 #define TY_SEG     0xE8
 #define TY_EXTERN  0xF0
@@ -823,6 +945,21 @@ struct dfmt {
 #define TYM_ELEMENTS(x) (((x) & 0xFFFFFF00) >> 8)
 
 #define TYS_ELEMENTS(x)  ((x) << 8)
+
+/*
+ * -----
+ * Special tokens
+ * -----
+ */
+
+enum special_tokens {
+    SPECIAL_ENUM_START = PREFIX_ENUM_LIMIT,
+    S_ABS = SPECIAL_ENUM_START,
+    S_BYTE, S_DWORD, S_FAR, S_LONG, S_NEAR, S_NOSPLIT,
+    S_OWORD, S_QWORD, S_REL, S_SHORT, S_STRICT, S_TO, S_TWORD, S_WORD,
+    SPECIAL_ENUM_LIMIT
+};
+
 /*
  * -----
  * Other
@@ -836,7 +973,11 @@ struct dfmt {
 
 #define elements(x)     ( sizeof(x) / sizeof(*(x)) )
 
-extern int tasm_compatible_mode;
+/*
+ * -----
+ * Global modes
+ * -----
+ */
 
 /*
  * This declaration passes the "pass" number to all other modules
@@ -846,7 +987,12 @@ extern int tasm_compatible_mode;
  *       2 = pass 2
  */
 
-extern int pass0;	/* this is globally known */
+extern int pass0;
+
+extern bool tasm_compatible_mode;
 extern int optimizing;
+extern int globalbits;          /* 16, 32 or 64-bit mode */
+extern int globalrel;		/* default to relative addressing? */
+extern int maxbits;		/* max bits supported by output */
 
 #endif
