@@ -36,12 +36,15 @@
  *
  * David Lindauer, LADsoft
  */
+#include "compiler.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
 #include <stdarg.h>             /* Note: we need the ANSI version of stdarg.h */
 #include <ctype.h>
+#include <inttypes.h>
 
 #include "nasm.h"
 #include "nasmlib.h"
@@ -57,7 +60,7 @@ static int ieee_uppercase;
 static efunc error;
 static ldfunc deflabel;
 static FILE *ofp;
-static int any_segs;
+static bool any_segs;
 static int arrindex;
 
 #define HUNKSIZE 1024           /* Size of the data hunk */
@@ -69,14 +72,14 @@ struct ieeeSection;
 struct LineNumber {
     struct LineNumber *next;
     struct ieeeSection *segment;
-    long offset;
-    long lineno;
+    int32_t offset;
+    int32_t lineno;
 };
 
 static struct FileName {
     struct FileName *next;
     char *name;
-    long index;
+    int32_t index;
 } *fnhead, **fntail;
 
 static struct Array {
@@ -88,16 +91,16 @@ static struct Array {
 static struct ieeePublic {
     struct ieeePublic *next;
     char *name;
-    long offset;
-    long segment;               /* only if it's far-absolute */
-    long index;
+    int32_t offset;
+    int32_t segment;               /* only if it's far-absolute */
+    int32_t index;
     int type;                   /* for debug purposes */
 } *fpubhead, **fpubtail, *last_defined;
 
 static struct ieeeExternal {
     struct ieeeExternal *next;
     char *name;
-    long commonsize;
+    int32_t commonsize;
 } *exthead, **exttail;
 
 static int externals;
@@ -112,24 +115,24 @@ static struct ieeeSection {
     struct ieeeObjData *data, *datacurr;
     struct ieeeSection *next;
     struct ieeeFixupp *fptr, *flptr;
-    long index;                 /* the NASM segment id */
-    long ieee_index;            /* the OBJ-file segment index */
-    long currentpos;
-    long align;                 /* can be SEG_ABS + absolute addr */
-    long startpos;
+    int32_t index;                 /* the NASM segment id */
+    int32_t ieee_index;            /* the OBJ-file segment index */
+    int32_t currentpos;
+    int32_t align;                 /* can be SEG_ABS + absolute addr */
+    int32_t startpos;
     enum {
         CMB_PRIVATE = 0,
         CMB_PUBLIC = 2,
         CMB_COMMON = 6
     } combine;
-    long use32;                 /* is this segment 32-bit? */
+    int32_t use32;                 /* is this segment 32-bit? */
     struct ieeePublic *pubhead, **pubtail, *lochead, **loctail;
     char *name;
 } *seghead, **segtail, *ieee_seg_needs_update;
 
 struct ieeeObjData {
     struct ieeeObjData *next;
-    unsigned char data[HUNKSIZE];
+    uint8_t data[HUNKSIZE];
 };
 
 struct ieeeFixupp {
@@ -144,35 +147,35 @@ struct ieeeFixupp {
         FT_EXTWRT = 6,
         FT_EXTSEG = 7
     } ftype;
-    short size;
-    long id1;
-    long id2;
-    long offset;
-    long addend;
+    int16_t size;
+    int32_t id1;
+    int32_t id2;
+    int32_t offset;
+    int32_t addend;
 };
 
-static long ieee_entry_seg, ieee_entry_ofs;
+static int32_t ieee_entry_seg, ieee_entry_ofs;
 static int checksum;
 
 extern struct ofmt of_ieee;
 
 static void ieee_data_new(struct ieeeSection *);
-static void ieee_write_fixup(long, long, struct ieeeSection *,
-                             int, unsigned long, long);
+static void ieee_write_fixup(int32_t, int32_t, struct ieeeSection *,
+                             int, uint32_t, int32_t);
 static void ieee_install_fixup(struct ieeeSection *, struct ieeeFixupp *);
-static long ieee_segment(char *, int, int *);
+static int32_t ieee_segment(char *, int, int *);
 static void ieee_write_file(int debuginfo);
 static void ieee_write_byte(struct ieeeSection *, int);
 static void ieee_write_word(struct ieeeSection *, int);
-static void ieee_write_dword(struct ieeeSection *, long);
+static void ieee_write_dword(struct ieeeSection *, int32_t);
 static void ieee_putascii(char *, ...);
 static void ieee_putcs(int);
-static long ieee_putld(long, long, unsigned char *);
-static long ieee_putlr(struct ieeeFixupp *);
+static int32_t ieee_putld(int32_t, int32_t, uint8_t *);
+static int32_t ieee_putlr(struct ieeeFixupp *);
 static void ieee_unqualified_name(char *, char *);
 
-/* 
- * pup init 
+/*
+ * pup init
  */
 static void ieee_init(FILE * fp, efunc errfunc, ldfunc ldef, evalfunc eval)
 {
@@ -180,7 +183,7 @@ static void ieee_init(FILE * fp, efunc errfunc, ldfunc ldef, evalfunc eval)
     ofp = fp;
     error = errfunc;
     deflabel = ldef;
-    any_segs = FALSE;
+    any_segs = false;
     fpubhead = NULL;
     fpubtail = &fpubhead;
     exthead = NULL;
@@ -191,7 +194,7 @@ static void ieee_init(FILE * fp, efunc errfunc, ldfunc ldef, evalfunc eval)
     seghead = ieee_seg_needs_update = NULL;
     segtail = &seghead;
     ieee_entry_seg = NO_SEG;
-    ieee_uppercase = FALSE;
+    ieee_uppercase = false;
     checksum = 0;
     of_ieee.current_dfmt->init(&of_ieee, NULL, fp, errfunc);
 }
@@ -251,8 +254,8 @@ static void ieee_cleanup(int debuginfo)
 /*
  * callback for labels
  */
-static void ieee_deflabel(char *name, long segment,
-                          long offset, int is_global, char *special)
+static void ieee_deflabel(char *name, int32_t segment,
+                          int32_t offset, int is_global, char *special)
 {
     /*
      * We have three cases:
@@ -263,7 +266,7 @@ static void ieee_deflabel(char *name, long segment,
      *
      * (ii) `segment' is one of our segments, or a SEG_ABS segment.
      * Save the label position for later output of a PUBDEF record.
-     * 
+     *
      *
      * (iii) `segment' is not one of our segments. Save the label
      * position for later output of an EXTDEF.
@@ -369,12 +372,12 @@ static void ieee_deflabel(char *name, long segment,
 /*
  * Put data out
  */
-static void ieee_out(long segto, const void *data, unsigned long type,
-                     long segment, long wrt)
+static void ieee_out(int32_t segto, const void *data, uint32_t type,
+                     int32_t segment, int32_t wrt)
 {
-    unsigned long size, realtype;
-    const unsigned char *ucdata;
-    long ldata;
+    uint32_t size, realtype;
+    const uint8_t *ucdata;
+    int32_t ldata;
     struct ieeeSection *seg;
 
     /*
@@ -388,7 +391,7 @@ static void ieee_out(long segto, const void *data, unsigned long type,
     }
 
     /*
-     * If `any_segs' is still FALSE, we must define a default
+     * If `any_segs' is still false, we must define a default
      * segment.
      */
     if (!any_segs) {
@@ -417,7 +420,7 @@ static void ieee_out(long segto, const void *data, unsigned long type,
         if (segment == NO_SEG && realtype != OUT_ADDRESS)
             error(ERR_NONFATAL, "relative call to absolute address not"
                   " supported by IEEE format");
-        ldata = *(long *)data;
+        ldata = *(int32_t *)data;
         if (realtype == OUT_REL2ADR)
             ldata += (size - 2);
         if (realtype == OUT_REL4ADR)
@@ -450,9 +453,9 @@ static void ieee_data_new(struct ieeeSection *segto)
  * but I might as well see what it is like on a harmless program.
  * If anyone wants to optimize this is a good canditate!
  */
-static void ieee_write_fixup(long segment, long wrt,
+static void ieee_write_fixup(int32_t segment, int32_t wrt,
                              struct ieeeSection *segto, int size,
-                             unsigned long realtype, long offset)
+                             uint32_t realtype, int32_t offset)
 {
     struct ieeeSection *target;
     struct ieeeFixupp s;
@@ -489,7 +492,7 @@ static void ieee_write_fixup(long segment, long wrt,
                          * Now we assume the segment field is being used
                          * to hold an extern index
                          */
-                        long i = segment / 2;
+                        int32_t i = segment / 2;
                         struct ExtBack *eb = ebhead;
                         while (i > EXT_BLKSIZ) {
                             if (eb)
@@ -539,7 +542,7 @@ static void ieee_write_fixup(long segment, long wrt,
                  * Now we assume the segment field is being used
                  * to hold an extern index
                  */
-                long i = segment / 2;
+                int32_t i = segment / 2;
                 struct ExtBack *eb = ebhead;
                 while (i > EXT_BLKSIZ) {
                     if (eb)
@@ -593,7 +596,7 @@ static void ieee_write_fixup(long segment, long wrt,
                  * Now we assume the segment field is being used
                  * to hold an extern index
                  */
-                long i = segment / 2;
+                int32_t i = segment / 2;
                 struct ExtBack *eb = ebhead;
                 while (i > EXT_BLKSIZ) {
                     if (eb)
@@ -650,7 +653,7 @@ static void ieee_install_fixup(struct ieeeSection *seg,
 /*
  * segment registry
  */
-static long ieee_segment(char *name, int pass, int *bits)
+static int32_t ieee_segment(char *name, int pass, int *bits)
 {
     /*
      * We call the label manager here to define a name for the new
@@ -666,7 +669,8 @@ static long ieee_segment(char *name, int pass, int *bits)
         return seghead->index;
     } else {
         struct ieeeSection *seg;
-        int ieee_idx, attrs, rn_error;
+        int ieee_idx, attrs;
+	bool rn_error;
         char *p;
 
         /*
@@ -715,7 +719,7 @@ static long ieee_segment(char *name, int pass, int *bits)
         segtail = &seg->next;
         seg->index = seg_alloc();
         seg->ieee_index = ieee_idx;
-        any_segs = TRUE;
+        any_segs = true;
         seg->name = NULL;
         seg->currentpos = 0;
         seg->align = 1;         /* default */
@@ -747,9 +751,9 @@ static long ieee_segment(char *name, int pass, int *bits)
             else if (!nasm_stricmp(p, "common"))
                 seg->combine = CMB_COMMON;
             else if (!nasm_stricmp(p, "use16"))
-                seg->use32 = FALSE;
+                seg->use32 = false;
             else if (!nasm_stricmp(p, "use32"))
-                seg->use32 = TRUE;
+                seg->use32 = true;
             else if (!nasm_strnicmp(p, "align=", 6)) {
                 seg->align = readnum(p + 6, &rn_error);
                 if (seg->align == 0)
@@ -787,10 +791,10 @@ static long ieee_segment(char *name, int pass, int *bits)
         ieee_seg_needs_update = seg;
         if (seg->align >= SEG_ABS)
             deflabel(name, NO_SEG, seg->align - SEG_ABS,
-                     NULL, FALSE, FALSE, &of_ieee, error);
+                     NULL, false, false, &of_ieee, error);
         else
             deflabel(name, seg->index + 1, 0L,
-                     NULL, FALSE, FALSE, &of_ieee, error);
+                     NULL, false, false, &of_ieee, error);
         ieee_seg_needs_update = NULL;
 
         if (seg->use32)
@@ -810,7 +814,7 @@ static int ieee_directive(char *directive, char *value, int pass)
     (void)value;
     (void)pass;
     if (!strcmp(directive, "uppercase")) {
-        ieee_uppercase = TRUE;
+        ieee_uppercase = true;
         return 1;
     }
     return 0;
@@ -819,7 +823,7 @@ static int ieee_directive(char *directive, char *value, int pass)
 /*
  * Return segment data
  */
-static long ieee_segbase(long segment)
+static int32_t ieee_segbase(int32_t segment)
 {
     struct ieeeSection *seg;
 
@@ -872,7 +876,7 @@ static void ieee_write_file(int debuginfo)
      */
     ieee_putascii("CO0,%02X%s.\r\n", strlen(boast), boast);
 
-    /* 
+    /*
      * write processor-specific information
      */
     ieee_putascii("AD8,4,L.\r\n");
@@ -886,7 +890,7 @@ static void ieee_write_file(int debuginfo)
                   1900 + thetime->tm_year, thetime->tm_mon + 1,
                   thetime->tm_mday, thetime->tm_hour, thetime->tm_min,
                   thetime->tm_sec);
-    /* 
+    /*
      * if debugging, dump file names
      */
     for (fn = fnhead; fn && debuginfo; fn = fn->next) {
@@ -898,9 +902,9 @@ static void ieee_write_file(int debuginfo)
      * the standard doesn't specify when to put checksums,
      * we'll just do it periodically.
      */
-    ieee_putcs(FALSE);
+    ieee_putcs(false);
 
-    /* 
+    /*
      * Write the section headers
      */
     seg = seghead;
@@ -950,7 +954,7 @@ static void ieee_write_file(int debuginfo)
                           ieee_entry_ofs);
     }
 
-    ieee_putcs(FALSE);
+    ieee_putcs(false);
     /*
      * Write the publics
      */
@@ -1007,7 +1011,7 @@ static void ieee_write_file(int debuginfo)
         ieee_putascii("NX%X,%02X%s.\r\n", i++, strlen(buf), buf);
         ext = ext->next;
     }
-    ieee_putcs(FALSE);
+    ieee_putcs(false);
 
     /*
      * IEEE doesn't have a standard pass break record
@@ -1056,7 +1060,7 @@ static void ieee_write_file(int debuginfo)
         seg = seg->next;
     while (seg) {
         if (seg->currentpos) {
-            long size, org = 0;
+            int32_t size, org = 0;
             data = seg->data;
             ieee_putascii("SB%X.\r\n", seg->ieee_index);
             fix = seg->fptr;
@@ -1081,7 +1085,7 @@ static void ieee_write_file(int debuginfo)
                 org = ieee_putld(org, org + size, data->data);
                 data = data->next;
             }
-            ieee_putcs(FALSE);
+            ieee_putcs(false);
 
         }
         seg = seg->next;
@@ -1106,7 +1110,7 @@ static void ieee_write_word(struct ieeeSection *seg, int data)
     ieee_write_byte(seg, (data >> 8) & 0xFF);
 }
 
-static void ieee_write_dword(struct ieeeSection *seg, long data)
+static void ieee_write_dword(struct ieeeSection *seg, int32_t data)
 {
     ieee_write_byte(seg, data & 0xFF);
     ieee_write_byte(seg, (data >> 8) & 0xFF);
@@ -1120,7 +1124,7 @@ static void ieee_putascii(char *format, ...)
     va_list ap;
 
     va_start(ap, format);
-    vsprintf(buffer, format, ap);
+    vsnprintf(buffer, sizeof(buffer), format, ap);
     l = strlen(buffer);
     for (i = 0; i < l; i++)
         if ((buffer[i] & 0xff) > 31)
@@ -1143,9 +1147,9 @@ static void ieee_putcs(int toclear)
     checksum = 0;
 }
 
-static long ieee_putld(long start, long end, unsigned char *buf)
+static int32_t ieee_putld(int32_t start, int32_t end, uint8_t *buf)
 {
-    long val;
+    int32_t val;
     if (start == end)
         return (start);
     val = start % HUNKSIZE;
@@ -1171,7 +1175,7 @@ static long ieee_putld(long start, long end, unsigned char *buf)
     ieee_putascii(".\r\n");
     return (start);
 }
-static long ieee_putlr(struct ieeeFixupp *p)
+static int32_t ieee_putlr(struct ieeeFixupp *p)
 {
 /*
  * To deal with the vagaries of segmentation the LADsoft linker
@@ -1189,51 +1193,51 @@ static long ieee_putlr(struct ieeeFixupp *p)
  * get away without defining the virtual mode...
  */
     char buf[40];
-    long size = p->size;
+    int32_t size = p->size;
     switch (p->ftype) {
     case FT_SEG:
         if (p->id1 < 0)
-            sprintf(buf, "%lX", -p->id1);
+            sprintf(buf, "%"PRIX32"", -p->id1);
         else
-            sprintf(buf, "L%lX,10,/", p->id1);
+            sprintf(buf, "L%"PRIX32",10,/", p->id1);
         break;
     case FT_OFS:
-        sprintf(buf, "R%lX,%lX,+", p->id1, p->addend);
+        sprintf(buf, "R%"PRIX32",%"PRIX32",+", p->id1, p->addend);
         break;
     case FT_REL:
-        sprintf(buf, "R%lX,%lX,+,P,-,%X,-", p->id1, p->addend, p->size);
+        sprintf(buf, "R%"PRIX32",%"PRIX32",+,P,-,%X,-", p->id1, p->addend, p->size);
         break;
 
     case FT_WRT:
         if (p->id2 < 0)
-            sprintf(buf, "R%lX,%lX,+,L%lX,+,%lX,-", p->id2, p->addend,
+            sprintf(buf, "R%"PRIX32",%"PRIX32",+,L%"PRIX32",+,%"PRIX32",-", p->id2, p->addend,
                     p->id2, -p->id1 * 16);
         else
-            sprintf(buf, "R%lX,%lX,+,L%lX,+,L%lX,-", p->id2, p->addend,
+            sprintf(buf, "R%"PRIX32",%"PRIX32",+,L%"PRIX32",+,L%"PRIX32",-", p->id2, p->addend,
                     p->id2, p->id1);
         break;
     case FT_EXT:
-        sprintf(buf, "X%lX", p->id1);
+        sprintf(buf, "X%"PRIX32"", p->id1);
         break;
     case FT_EXTREL:
-        sprintf(buf, "X%lX,P,-,%lX,-", p->id1, size);
+        sprintf(buf, "X%"PRIX32",P,-,%"PRIX32",-", p->id1, size);
         break;
     case FT_EXTSEG:
         /* We needed a non-ieee hack here.
          * We introduce the Y variable, which is the low
          * limit of the native segment the extern resides in
          */
-        sprintf(buf, "Y%lX,10,/", p->id1);
+        sprintf(buf, "Y%"PRIX32",10,/", p->id1);
         break;
     case FT_EXTWRT:
         if (p->id2 < 0)
-            sprintf(buf, "X%lX,Y%lX,+,%lX,-", p->id2, p->id2,
+            sprintf(buf, "X%"PRIX32",Y%"PRIX32",+,%"PRIX32",-", p->id2, p->id2,
                     -p->id1 * 16);
         else
-            sprintf(buf, "X%lX,Y%lX,+,L%lX,-", p->id2, p->id2, p->id1);
+            sprintf(buf, "X%"PRIX32",Y%"PRIX32",+,L%"PRIX32",-", p->id2, p->id2, p->id1);
         break;
     }
-    ieee_putascii("LR(%s,%lX).\r\n", buf, size);
+    ieee_putascii("LR(%s,%"PRIX32").\r\n", buf, size);
 
     return (size);
 }
@@ -1263,7 +1267,7 @@ void dbgls_init(struct ofmt *of, void *id, FILE * fp, efunc error)
     arrhead = NULL;
     arrtail = &arrhead;
     ieee_segment("??LINE", 2, &tempint);
-    any_segs = FALSE;
+    any_segs = false;
 }
 static void dbgls_cleanup(void)
 {
@@ -1296,7 +1300,7 @@ static void dbgls_cleanup(void)
  * so, we have to make sure the ??LINE segment is avaialbe
  * as the first segment when this debug format is selected
  */
-static void dbgls_linnum(const char *lnfname, long lineno, long segto)
+static void dbgls_linnum(const char *lnfname, int32_t lineno, int32_t segto)
 {
     struct FileName *fn;
     struct ieeeSection *seg;
@@ -1305,7 +1309,7 @@ static void dbgls_linnum(const char *lnfname, long lineno, long segto)
         return;
 
     /*
-     * If `any_segs' is still FALSE, we must define a default
+     * If `any_segs' is still false, we must define a default
      * segment.
      */
     if (!any_segs) {
@@ -1343,14 +1347,13 @@ static void dbgls_linnum(const char *lnfname, long lineno, long segto)
                      seg->currentpos);
 
 }
-static void dbgls_deflabel(char *name, long segment,
-                           long offset, int is_global, char *special)
+static void dbgls_deflabel(char *name, int32_t segment,
+                           int32_t offset, int is_global, char *special)
 {
     struct ieeeSection *seg;
-    int used_special;           /* have we used the special text? */
 
-    /* Keep compiler from warning about special and used_special */
-    used_special = special ? FALSE : FALSE;
+    /* Keep compiler from warning about special */
+    (void)special;
 
     /*
      * If it's a special-retry from pass two, discard it.
@@ -1379,7 +1382,7 @@ static void dbgls_deflabel(char *name, long segment,
     }
 
     /*
-     * If `any_segs' is still FALSE, we might need to define a
+     * If `any_segs' is still false, we might need to define a
      * default segment, if they're trying to declare a label in
      * `first_seg'.  But the label should exist due to a prior
      * call to ieee_deflabel so we can skip that.
@@ -1403,7 +1406,7 @@ static void dbgls_deflabel(char *name, long segment,
             }
         }
 }
-static void dbgls_typevalue(long type)
+static void dbgls_typevalue(int32_t type)
 {
     int elem = TYM_ELEMENTS(type);
     type = TYM_TYPE(type);
@@ -1413,7 +1416,7 @@ static void dbgls_typevalue(long type)
 
     switch (type) {
     case TY_BYTE:
-        last_defined->type = 1; /* unsigned char */
+        last_defined->type = 1; /* uint8_t */
         break;
     case TY_WORD:
         last_defined->type = 3; /* unsigned word */
