@@ -10,8 +10,9 @@
 #ifndef NASM_PARSER_H
 #define NASM_PARSER_H
 
-insn *parse_line (int pass, char *buffer, insn *result,
-		  efunc error, evalfunc evaluate, evalinfofunc einfo);
-void cleanup_insn (insn *instruction);
+void parser_global_info(struct ofmt *output, struct location * locp);
+insn *parse_line(int pass, char *buffer, insn * result,
+                 efunc error, evalfunc evaluate, ldfunc ldef);
+void cleanup_insn(insn * instruction);
 
 #endif
